export default class AST_RENDERER {

    constructor(oTemplates, oGlobalContext) {
        
        this.oTemplates = oTemplates;
        this.oGlobalContext = oGlobalContext;

        this.mTemplates = new Map();

        // Will be set when constructed
        this.oData = null;
        this.vTempData = null;

        this.registerTemplates();
    }

    registerTemplates(oTemplates) {

        // Loop through all predefined templates
        for (let oTemplate of oTemplates) {

            if (!this.mTemplates.has(oTemplate.name)) {
    
                this.mTemplates.set(oTemplate.name, oTemplate.contents);
            }
            else {
    
                console.error(`Attempt to redefine template: ${oTemplate.name} was made.`);
            }
        }


    }

    convertObjectToString(vObject) {

        let sResolvedValue = null;

        if (Array.isArray(vObject)) {

            sResolvedValue = "[Array]";
        }
        else if (vObject === null) {

            sResolvedValue = "[Null]";
        }
        else {

            sResolvedValue = "[Object]";
        }

        return sResolvedValue;
    }
  
    resolveLogic(oLogic) {

        const F_EXTRACT_EXP_VALUE = (oExpression) => {

            // First evaluated eval
            switch (oExpression.eval.type) {

                case "literal":
                    
                    return oExpression.eval.value;

                case "data":

                    return this.resolveData(oExpression.eval, false);

                default:

                    console.error("Unknown expression type: ${oExpression.eval.value}");

            }


        }

        const F_RESOLVE_CONDITION = (oCondition) => {

            
            if (oCondition.type === "unary") {
    
                let vExpressValue = F_EXTRACT_EXP_VALUE(oCondition.expression);

                if (oCondition.expression.inverse) {

                    return (vExpressValue) ? false: true;
                }
                else {

                    return (vExpressValue) ? true : false;
                }

            }
            else if (oCondition.type === "binary") {

                let vLeftEval = F_EXTRACT_EXP_VALUE(oCondition.expression.left);
                let vRightEval = F_EXTRACT_EXP_VALUE(oCondition.expression.right);

                // We will assume the result is false and toggle to true if needed!
                let bResult = false;

                switch (oCondition.expression.op) {

                    case "==":

                        if (vLeftEval == vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "===":

                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "!=":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case "!==":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case ">":

                        if (vLeftEval > vRightEval) {
                            bResult = true;
                        } 
                        break;
                        
                    case ">=":

                        if (vLeftEval >= vRightEval) {
                            bResult = true;
                        } 
                        break;

                    case "<":

                        if (vLeftEval < vRightEval) {
                            bResult = true;
                        } 
                        break;


                    case "<=":

                        if (vLeftEval <= vRightEval) {
                            bResult = true;
                        } 
                        break;

                }

                return bResult;
            }

        }

        // Loop through all of the supplied conditionals, as 
        for (let oConditional of oLogic.conditionals) {

            if (F_RESOLVE_CONDITION(oConditional.conditional)) {
                return oConditional.consequence
            }

        }

        if (oLogic.alternative && oLogic.alternative.length) {

            return oLogic.alternative;
        }

        return null;
    }

    resolveData(oNode, bError = true) {

        // Continue only if the contents node for data has a string with length
        if (oNode && typeof oNode.value === "string" && oNode.value.length) {

            let sPath = oNode.value.split('.');

            if (sPath[0] === "this") {

                if (!this.oData.this) {
                    sPath = sPath.slice(1);
                }
            }

            let oContext = null;

            // Determine the "data" context
            switch (sPath[0]) {

                case "_global":

                    oContext = this.oGlobalContext;
                    sPath = sPath.slice(1);
                    break;

                case "_local":

                    oContext = this.oLocalContext;
                    sPath = sPath.slice(1);
                    break;

                case "_data":

                    oContext = this.oData;
                    sPath = sPath.slice(1);
                    break;

                default:

                    oContext = this.oData;
                    break;
            }   

            // Loop through the paths
            for (let p = 0, pLen = sPath.length; p < pLen; p++) {

                // Check to see if this part of the path exists!
                if (oContext.hasOwnProperty(sPath[p])) {

                    if (p + 1 === pLen) {

                        return oContext[sPath[p]];
                    }
                    else {

                        oContext = oContext[sPath[p]];
                        continue;
                    }

                }
                else {

                    if (bError) {
                        console.error(`Requested invalid data context path: ${oNode.value}`);
                    }

                    return null;
                }

            }

        }

    }

    generateAttributes(aoAttributes) {

        let oAttributes = {};

        const F_HANDLE_LITERAL_ATTR = (oLiteralNode, sAttrName) => {

            switch (typeof oLiteralNode.value) {

                case "string":
                    oAttributes[sAttrName].push(oLiteralNode.value);
                    break;

                case "number":
                case "boolean":
                    oAttributes[osAttrName].push(oLiteralNode.value.toString());
                    break;
            }

        }

        const F_HANDLE_DATA_ATTR = (oDataNode, sAttrName) => {

            let vDataValue = this.resolveData(oDataNode);

            switch(typeof vDataValue) {

                case "boolean":
                case "number":
                    vDataValue = vDataValue.toString();
                    break;
                
                case "object":

                    vDataValue = this.convertObjectToString(vDataValue);
                    break;

            }

            if (vDataValue) {

                oAttributes[sAttrName].push(vDataValue);
            }

        }

        const F_LOOP_ATTRIBUTES = (aoAttributes, sSpecificAttr) => {

            // Loop through attribute objects
            for (let oNode of aoAttributes) {
                
                // Shift through all the object types
                switch (oNode.type) {
        
                    // Handle static typed attributes, or those that have dynamics inside of the attribute
                    case "attr":
    
                        if (!oAttributes[oNode.name]) {
                            oAttributes[oNode.name] = [];
                        }
    
                        if (oNode.contents && oNode.contents.length) {
    
                            for (let av = 0, avLen = oNode.contents.length; av < avLen; av++) {
    
                                switch (oNode.contents[av].type) {
    
                                    case "literal":
    
                                        F_HANDLE_LITERAL_ATTR(oNode.contents[av], oNode.name);
                                        break
    
                                    case "data":
    
                                        F_HANDLE_DATA_ATTR(oNode.contents[av], oNode.name);
                                        break;
    
                                    case "logic":
                                    case "logical":

                                        console.log("logic in attribute");
    
                                        let aoLogicResolve = this.resolveLogic(oNode.contents[av]);
    
                                        if (aoLogicResolve && aoLogicResolve.length) {
                                            F_LOOP_ATTRIBUTES(aoLogicResolve, oNode.name);
                                        }
    
                                        break;
    
                                }
    
                            }
    
                        }
    
                        break;

                    case "logic":

                        let aoLogicResolve = this.resolveLogic(oNode);
    
                        if (aoLogicResolve && aoLogicResolve.length) {
                            F_LOOP_ATTRIBUTES(aoLogicResolve);
                        }

                        break;

                    default:

                        if (sSpecificAttr) {

                            if (!oAttributes[sSpecificAttr]) {
                                oAttributes[sSpecificAttr] = [];
                            }

                            switch(oNode.type) {

                                case "literal":
            
                                    F_HANDLE_LITERAL_ATTR(oNode, sSpecificAttr);
                                    break;

                                case "data":

                                    F_HANDLE_DATA_ATTR(oNode, sSpecificAttr);
                                    break;
                            }

                        }
    
                }
                
            }

            if (Object.keys(oAttributes).length) {

                return oAttributes;
            }

            return null;
        }

        // Cal the intial attributes list
        F_LOOP_ATTRIBUTES(aoAttributes);

        if (Object.keys(oAttributes).length) {

            return oAttributes;
        }

        return null;
    }

    subRenderer(oAdditionalTemplates) {

    }

    generate(aoSource, dParentNode, oLocalContext, oLocalTemplates) {

        // Only set the generate when its provided for the first time
        if (this.oData === null && oData) {
            this.oData = oData;
        }

        if (!dParentNode) {
            dParentNode = document.createDocumentFragment();
        }

        if (typeof aoSource === "object" && !Array.isArray(aoSource)) {
            aoSource = [ aoSource ];
        }
        else if (!Array.isArray(aoSource)) {

            throw new Error("Source AST is not an array!");
        }

        for (let a = 0, aLen = aoSource.length; a < aLen; a++) {

            let oNode = aoSource[a];

            switch (oNode.type) {

                case "tag":

                    // Call the create element function
                    let dNewElement = document.createElement(oNode.name);

                    // Check and resolve attribute objects 
                    if (oNode.attributes && Array.isArray(oNode.attributes) && oNode.attributes.length) {

                        let oFinalizedAttributes = this.generateAttributes(oNode.attributes);

                        if (oFinalizedAttributes) {

                            for (let sAttributeName in oFinalizedAttributes) {

                                switch(sAttributeName) {

                                    case "checked":
                                    case "mutli":
                                    case "selected":

                                        console.warn(`valueless attributes not supported`);
                                        break;

                                    default:

                                        dNewElement.setAttribute(sAttributeName, oFinalizedAttributes[sAttributeName].join(" ").trim());
                                        break;

                                }

                            }

                        }
                    }

                    // Check and render children
                    if (oNode.contents && Array.isArray(oNode.contents) && oNode.contents.length) {

                        let dChildren = this.generate(oNode.contents);

                        // Loop through all children and add them to the parent (dNewEelement)
                        while (dChildren.hasChildNodes()) {  

                            dNewElement.appendChild(dChildren.firstChild.cloneNode(true))

                            dChildren.removeChild(dChildren.firstChild);
                        }

                    }

                    dParentNode.appendChild(dNewElement);

                    break;

                case "literal":

                    switch(typeof oNode.value) {

                        case "string":

                            dParentNode.appendChild(document.createTextNode(`${oNode.value}`));
                            break;

                        case "object":

                            dParentNode.appendChild(document.createTextNode(this.convertObjectToString(oNode.vlaue)));
                            break;

                        case "number":
                        case "boolean":

                            dParentNode.appendChild(document.createTextNode(oNode.value.toString()));
                            break;
                    }


                    break;

                case "data":

                    // In this location, the value is being called after or inside of a tag node

                    let sResolvedValue = this.resolveData(oNode);

                    switch (typeof sResolvedValue) {

                        case "object":

                            sResolvedValue = this.convertObjectToString(sResolvedValue);
                            break;

                        case "boolean":
                        case "number":

                            sResolvedValue = sResolvedValue.toString();
                            break;

                    }

                    dParentNode.appendChild(document.createTextNode(sResolvedValue));

                    break;

                case "log":

                    for (let oContents of oNode.contents) {

                        let sLog = null;

                        switch (oContents.type) {

                            case "literal":

                                sLog = oContents.value;
                                break;
                                
                            case "data":
                            
                                sLog = this.resolveData(oNode);
                                break;

                            default:

                                console.error(`[PLATES LOG] => Only supports literal and data value, was passed a ${oNode.type} node.`);
                                break;
                                    
                        }

                        if (sLog !== null) {

                            console.log(`[PLATES LOG] => ${oContents.value}`);
                        }
                    }
                            

                    break;

                case "logic":

                    // Resolve the logic node
                    let aoLogicResolve = this.resolveLogic(oNode);
                    
                    // If anything is returned, process it with a internal generate
                    if (aoLogicResolve && aoLogicResolve.length) {
                        
                        this.generate(aoLogicResolve, dParentNode);

                    }
                    
                    break;

                case "loop":

                    let vDataSet = this.resolveData(oNode.source);

                    if (vDataSet !== null) {

                        switch (oNode.name) {

                            case "for":
    
                                if (typeof vDataSet === "object") {

                                    let iLoopIndex = 0;

                                    if (Array.isArray(vDataSet)) {
    
                                        for (let vData of vDataSet) {
    
                                            // Create a new instance, passinf only the templates, and contexts we know about
                                            let cLoopSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

                                            let oLoopData = {
                                                "@index": iLoopIndex,
                                                "@key": iLoopIndex,
                                                "@source": vDataSet,
                                                "this": vData
                                            }

                                            let dLoopChildren = cLoopSubRender.generate(oNode.contents, false, oLoopData);

                                            // Loop through all children and add them to the parent (dNewEelement)
                                            while (dLoopChildren.hasChildNodes()) {  

                                                dParentNode.appendChild(dLoopChildren.firstChild.cloneNode(true))

                                                dLoopChildren.removeChild(dLoopChildren.firstChild);
                                            }

                                            
                                            // Null out the template subrenderer
                                            cLoopSubRender = null;
                                            
                                            iLoopIndex += 1;
                                        }
    
                                    }
                                    else {
    
                                        for (let sDatasetKey in vDataSet) {

                                            // Create a new instance, passinf only the templates, and contexts we know about
                                            let cLoopSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

                                            let oLoopData = {
                                                "@index": iLoopIndex,
                                                "@key": sDatasetKey,
                                                "@source": vDataSet,
                                                "this": vDataSet[sDatasetKey],
                                            }

                                            console.log("Object loop");

                                            console.log(oLoopData);

                                            let dLoopChildren = cLoopSubRender.generate(oNode.contents, false, oLoopData);

                                            // Loop through all children and add them to the parent (dNewEelement)
                                            while (dLoopChildren.hasChildNodes()) {  

                                                dParentNode.appendChild(dLoopChildren.firstChild.cloneNode(true))

                                                dLoopChildren.removeChild(dLoopChildren.firstChild);
                                            }

                                            
                                            // Null out the template subrenderer
                                            cLoopSubRender = null;
                                            
                                            iLoopIndex += 1;

                                        }
                                        
                                    }
    
                                }
                                else {
    
                                    console.error(`Data source is not an interable: ${vDataSet}`);
                                }
    
                                break;
    
                        }

                    }

                    break;

                case "template":

                    // verify this template is real first
                    if (this.mTemplates.has(oNode.name)) {

                        // Create a new instance, passinf only the templates, and contexts we know about
                        let cSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

                        if (oNode.arguments && oNode.arguments.length) {

                            let oNewData = {};

                            for (let arg of oNode.arguments) {

                                if (typeof arg === "object") {

                                    switch (arg.type) {

                                        case "data":

                                            let vData = this.resolveData(arg);

                                            if (typeof vData === "object" && !Array.isArray(vData)) {

                                                oNewData = {
                                                    ...oNewData,
                                                    ...vData
                                                };

                                            }
                                            else {

                                                console.log(`Template arguments can only be references to another object or a statically asigned key value pair where the value can be any data type.`)
                                            }

                                            break;

                                        case "keyValue":

                                            let sKey = arg.key;
                                            let vKeyValue = null;

                                            for (let oStaticContents of arg.contents) {

                                                switch(oStaticContents.type) {

                                                    case "literal":
                                                    
                                                        if (vKeyValue) {
                                                            vKeyValue += oStaticContents.value;
                                                        }
                                                        else {

                                                            vKeyValue = oStaticContents.value;
                                                        }

                                                        break;

                                                    case "data":

                                                        let vData = this.resolveData(oStaticContents);

                                                        if (vKeyValue) {

                                                            vKeyValue += vData;
                                                        }
                                                        else {

                                                            vKeyValue = vData;
                                                        }

                                                        break;

                                                }
                                                
                                            }

                                            oNewData[sKey] = vKeyValue;
                                            break;
                                    }

                                }

                            }

                            let dSubChildren = cSubRender.generate(this.mTemplates.get(oNode.name), false, oNewData);

                            // Loop through all children and add them to the parent (dNewEelement)
                            while (dSubChildren.hasChildNodes()) {  

                                dParentNode.appendChild(dSubChildren.firstChild.cloneNode(true))

                                dSubChildren.removeChild(dSubChildren.firstChild);
                            }

                            
                            // Null out the template subrenderer
                            cSubRender = null;
                        }
                        else {

                            this.generate(this.mTemplates.get(oNode.name), dParentNode);
                        }
                        
                    }
                    else {

                        console.error(`Requested unknown template: ${oNode.name}`);
                    }

                    break;

                default:

                    console.error(`Unknown tag was discovered: ${oNode.type}`);
                    break;
            }

        }

        if (this.vTempData !== null) {
            this.vTempData = null;
        }

        return dParentNode;

    }
}