
import * as DOM from './dom.js'
import DATA_INTERFACE from './dataInterface';

/*

Naming of methods:
    Most methods are not designd to be overritten, which is why they are prefixed with an '_'. Instead they call child methods if defined after running necessary lifecycle code. 

Screen lifecycle:
    After each parent lifecycle method call, the child method (function without an underscore) is called. At this point the only method that requires a 'super' call on the child method is the constructor. 

        1) constructor
        2) _init
        3) _preRender (No child method call allowed)
        4) _render (No child method call allowed)
        5) _postRender

Special methods:

    _unload: Called when a screen is unloaded due to nav to or route functions. 


Screen schema life cycle: 

- Base schema (defined in file or default set in screen).
- dataKeys are pulled in and populate screenSchema object from template if present. 
- Screen Schema is sent to data manager to retreive all necessary values. 
- Data manager sends populated screenSchema back to page. 
- View key data is merged into screen schema (can happen before data retreival).
- Use populated screenSchema to populate fields in screen.

*/

export default class SCREEN {

    constructor(oProps) {

        this.sTitle = null;

        this.oParentRoute = (oProps.oParentRoute) ? oProps.oParentRoute : null;
        this.oNavState = (oProps.oNavState) ? oProps.oNavState : null;
        
        this.oHistoryState = (this.oNavState && this.oNavState.oHistoryState) ? this.oNavState.oHistoryState : null;
        this.oScreenState = null;

        this._dScreenContainer = (this.oParentRoute.screenContainer) ? this.oParentRoute.screenContainer : cAPP.appBody;   

        this.dStyleContainer = document.createElement('style');
        this.dViewContainer = document.createElement('div');        

        this.cDataInterface = new DATA_INTERFACE();
        this.cDataInterface.init();

        this.oLifeCycleNames = {
            init: 'init',
            reload:'reload',
            preRender:'preRender',
            render:'render',
            postRender:'postRender',
            unload:'unload',            
        }

        this._sGuardCycle = this.oLifeCycleNames.init;

        // Stub out commonly referenced elements as a reference. 
        this.sScreenStyles = null;
        this.sScreenTemplate = null;
        
        this.sRenderTemplate = null;
        this.cRenderer = null;
        this.oRenderData = {};

        this.dScreenTemplateFragment = null;
        this.oPageSchemaEndpoint = null;
        this.aTemplateFields = null;
    }

    set sGuardCycle(sCycle){
        if(sCycle && this.oLifeCycleNames.hasOwnProperty(sCycle)){
            this._sGuardCycle = sCycle;
        }
    }

    //================
    //Lifecycle methods
    //================
    _navGuard(){
        if(this.navGuard){
            let navGuardResponse = this.navGuard();
            return navGuardResponse;
        }
        return false;
    }

    _guard(){
        /*DEBUG: START*/
        console.log("SCREEN - _guard - "+ this.sTitle);
        /*DEBUG: END*/
        
        if(this.guard){
            let guardResponse = this.guard();   
            
            if(guardResponse === true){
                //TODO: Set default action when child guard method returns true. 
                /*DEBUG: START*/
                console.log('Guard tripped - ABORT');                
                /*DEBUG: END*/
            }
            else if(typeof guardResponse == "string"){    // if string - nav there
                cAPP.navTo(guardResponse);                
            }
        }     
    }

    _checkGuard(sCycle){
        if(sCycle == this._sGuardCycle){
            this._guard();
        }
    }

    _init () {
        /*DEBUG: START*/
        console.log("SCREEN - _init - "+ this.sTitle);
        /*DEBUG: END*/

        DOM.clearElements(this._dScreenContainer);

        this._dScreenContainer.appendChild(this.dViewContainer);        
        this._setScreenStyles();
    
        // Call child method and next lifecycle method
        this._callChildMethod(this.oLifeCycleNames.init).then(()=>{
            this._checkGuard(this.oLifeCycleNames.init);
            
            if(this.oHistoryState){
                this._reload();    
            }
            else{
                this._preRender();
            }            
        })  
          
    }

    _reload(){
        /*DEBUG: START*/
        console.log("SCREEN - _reload - "+ this.sTitle);
        /*DEBUG: END*/
        this._callChildMethod(this.oLifeCycleNames.reload).then(()=>{
            this._checkGuard(this.oLifeCycleNames.reload);
            this._preRender();
        }) 
    }

    _preRender(){
        /*DEBUG: START*/
        console.log("SCREEN - _preRender - "+ this.sTitle);
        /*DEBUG: END*/

        let sCurrentScreen = (this.oNavState.sDestination) ? this.oNavState.sDestination.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, "") : "";

        let sPreviousScreen = (this.oNavState.sSource) ? this.oNavState.sSource.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, ""): "";

        const E_SCREEN_LOAD = new CustomEvent('screenPreRender', { 
            bubble: true,
            detail: {
                sLoadingScreen: sCurrentScreen,
                sPreviousScreen: sPreviousScreen,
                oNavState: this.oNavState,
                dtStamp: Date.now() 
            }
        });

        this._dScreenContainer.dispatchEvent(E_SCREEN_LOAD);

        this._populateSchemaWithPageSchema().then(()=>{

            this._createScreenTemplateFragment(); 

            if(this.dScreenTemplateFragment){
                // Call updateLinks here as the content should be present in the screenFragment. 
                this._updateLinks();
                this._bindActions();

                this._addTemplateFieldsToScreenSchema();    
            }

            // Call next lifecycle method            
            // PreRender does not allow for child methods to be called. 
            this._checkGuard(this.oLifeCycleNames.preRender);
            this._render();            
        })
    }

    _render () { 
       /*DEBUG: START*/
        console.log("SCREEN - _render - "+ this.sTitle);
        /*DEBUG: END*/

        this._populateSchemaWithData().then(()=>{
            this._updateSchemaWithViewKeys();
            this._updateTemplateWithSchema();            
            this._renderScreenTemplateFragment()
            
            // Call next lifecycle method            
            // Render does not allow for child methods to be called. 
            this._checkGuard(this.oLifeCycleNames.render);
            this._postRender();            
        })
    }

    _postRender(){
        /*DEBUG: START*/
        console.log("SCREEN - _postRender - "+ this.sTitle);
        /*DEBUG: END*/

        // Emit a custom event - Used for Puppeteer build
        let sCurrentScreen = (this.oNavState.sDestination) ? this.oNavState.sDestination.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, "") : "";

        let sPreviousScreen = (this.oNavState.sSource) ? this.oNavState.sSource.replace(`${document.location.protocol}//${document.location.hostname}:${document.location.port}`, ""): "";

        const E_SCREEN_LOADED = new CustomEvent('screenPostRender', { 
            bubble: true,
            detail: {
                sLoadingScreen: sCurrentScreen,
                sPreviousScreen: sPreviousScreen,
                oNavState: this.oNavState,
                dtStamp: Date.now() 
            }
        });

        document.body.dispatchEvent(E_SCREEN_LOADED);  

        // Call child method and next lifecycle method
        this._callChildMethod(this.oLifeCycleNames.postRender).then(()=>{
            this._checkGuard(this.oLifeCycleNames.postRender);
        })
    }

    _unload(){
        /*DEBUG: START*/
        console.log('SCREEN - Unload - '+ this.sTitle);
        /*DEBUG: END*/
       
        if(this[this.oLifeCycleNames.unload] && typeof this[this.oLifeCycleNames.unload] == "function"){
            this[this.oLifeCycleNames.unload]();
        }

        return this.oScreenState;
    }

    
    //================
    // Other utility methods
    //================
    
    _setScreenStyles(){
        if(this.sScreenStyles){
            this.dStyleContainer.innerHTML = this.sScreenStyles;
            this._dScreenContainer.prepend(this.dStyleContainer);
        }
    }

    async _callChildMethod(sMethod){
        if(this[sMethod] && typeof this[sMethod] == "function"){
            
            await Promise.resolve( this[sMethod]() ).then( value => {
               return value;
            } );
        }
        else{
            return null;
        }
    }

    _renderScreenTemplateFragment(){
        if(this.dViewContainer){
            if(this.dScreenTemplateFragment){
                DOM.clearElements(this.dViewContainer);
                this.dViewContainer.appendChild(this.dScreenTemplateFragment);
            }
        }
    }

    _updateLinks(){
        let adLinks = this.dScreenTemplateFragment.querySelectorAll('a');

        for (let a = 0, aLen = adLinks.length; a < aLen; a++) {
            let dLink = adLinks[a];
            
            dLink.addEventListener('click', (evt) => {
                evt.preventDefault();
                let sNavTo = evt.target.getAttribute('href');
                cAPP.navTo(sNavTo, null, evt,);
            });
        }
    }

    _bindActions(){
        this.dActionContainer = this.dScreenTemplateFragment;
        //If building html in render function there wont be a screen template fragment.
        if(!this.dScreenTemplateFragment){
            this.dActionContainer = this.dViewContainer;
        }

        let aHasAction = this.dActionContainer.querySelectorAll("[data-action]");
        
        if(aHasAction && aHasAction.length > 0){
            for(let i =0; i < aHasAction.length; i++){
                let dHasAction = aHasAction[i];

                let sAction = dHasAction.dataset.action;
                let sActionType = dHasAction.dataset.actionType;

                if(sAction && sAction != "" && this[sAction] && sActionType && sActionType != "" && !dHasAction.onclick){
                    switch(sActionType){
                        case 'onclick':
                            //If defining a click event, should set type to button incase defaulted to submit. 
                            if(dHasAction.nodeName == "BUTTON"){
                                dHasAction.type = "button";    
                            }                            
                            dHasAction.onclick = this[sAction].bind(this);
                        break;                        
                    }
                }
            }
        }
    }  

    _registerService(sServiceName, oServiceProps){
        if(sServiceName && oServiceProps){

            if(!this.cDataInterface){
                this.cDataInterface = new DATA_INTERFACE();
                this.cDataInterface.init();
            }

            return this.cDataInterface.registerService(sServiceName, oServiceProps);
        }
        return null;
    }
 
    _registerDataStore(sDataStoreName, oDataStoreProps){
        if(sDataStoreName && oDataStoreProps){
            
            if(!this.cDataInterface){
                this.cDataInterface = new DATA_INTERFACE();
                this.cDataInterface.init();
            }

            return this.cDataInterface.registerDataStore(sDataStoreName, oDataStoreProps);
        }
        return null;
    }

    _addTemplateFieldsToScreenSchema(){
        if(!this.oScreenSchema){
            this.oScreenSchema = {};
        }

        let aTemplateFields = this.dScreenTemplateFragment.querySelectorAll("[dataKey], [viewKey], [source]");
     
        if(aTemplateFields && aTemplateFields.length > 0){
            for(let i =0; i < aTemplateFields.length; i++){
                let dTemplateField = aTemplateFields[i];
                let dTemplateFieldDataKey = dTemplateField.getAttribute('dataKey');
               
                // dataKey exists and is present in screen schema
                if(dTemplateFieldDataKey && this.oScreenSchema[dTemplateFieldDataKey]){
                    //Pull viewKey to screenSchema if it exists
                    if(dTemplateField.getAttribute("viewKey") && !this.oScreenSchema[dTemplateFieldDataKey].viewKey){
                        this.oScreenSchema[dTemplateFieldDataKey].viewKey = dTemplateField.getAttribute("viewKey");
                    }

                    //Pull source to screenSchema if it exists and is not present in schema
                    if(dTemplateField.getAttribute("source") && !this.oScreenSchema[dTemplateFieldDataKey].source){
                        this.oScreenSchema[dTemplateFieldDataKey].source = dTemplateField.getAttribute("source");
                    }
                }

                // dataKey exists but not present in screen schema. 
                else if(dTemplateFieldDataKey){
                   
                    // Pull all field attributes and save to screenSchema
                    let oDataProperties = {};
                    for (const [dataProp, value] of Object.entries(dTemplateField.dataset)) {
                        oDataProperties[dataProp] = value;
                    }   

                    this.oScreenSchema[dTemplateFieldDataKey] = {...oDataProperties};

                    //Pull viewKey to screenSchema if it exists
                    if(dTemplateField.getAttribute("viewKey")){
                        this.oScreenSchema[dTemplateFieldDataKey].viewKey = dTemplateField.getAttribute("viewKey");
                    }

                    //Pull source to screenSchema if it exists
                    if(dTemplateField.getAttribute("source")){
                        this.oScreenSchema[dTemplateFieldDataKey].source = dTemplateField.getAttribute("source");
                    }
                }

                // dataKey does not exist on field so will need to create unique ID and add to screenSchema. 
                else{
                    
                    let randomDataKey = "DK_" + i + "_" + Math.floor(Math.random() * 1000);
                    //Set physical attribute on web component
                    dTemplateField.setAttribute("dataKey", randomDataKey);

                    this.oScreenSchema[randomDataKey] = {};

                    //Pull viewKey to screenSchema if it exists
                    if(dTemplateField.getAttribute("viewKey")){
                        this.oScreenSchema[randomDataKey].viewKey = dTemplateField.getAttribute("viewKey");
                    }

                    //Pull source to screenSchema if it exists
                    if(dTemplateField.getAttribute("source")){
                        this.oScreenSchema[randomDataKey].source = dTemplateField.getAttribute("source");
                    }
                }
            }
           
            // Save reference to template fields to be used later in data binding. This eliminates the need to query the template fragment a second time. 
            this.aTemplateFields = aTemplateFields;

        }        
    }

    _addTemplateViewKeysToScreenSchema(){
        console.error("Should update this function call to _addTemplateFieldsToScreenSchema");
        if(!this.oScreenSchema){
            this.oScreenSchema = {};
        }

        let aHasDataSource = this.dScreenTemplateFragment.querySelectorAll("[source]");

        if(aHasDataSource && aHasDataSource.length > 0){
            for(let i =0; i < aHasDataSource.length; i++){
                let dHasDataSource = aHasDataSource[i];

                // Only add the field and properties to the screenSchema if it currently not present. 
                if(dHasDataSource.dataset.viewkey && !this.oScreenSchema[dHasDataSource.dataset.viewKey]){
                    
                    // Pull all field attributes and save to screenSchema
                    let oDataProperties = {};
                    for (const [dataProp, value] of Object.entries(dHasDataSource.dataset)) {
                        oDataProperties[dataProp] = value;
                    }   

                    this.oScreenSchema[dHasDataSource.dataset.viewkey] = {...oDataProperties,...this.oScreenSchema[dHasDataSource.dataset.viewkey]};
                }
            }
        }        
    }

    _updateSchemaWithViewKeys(){
        if(this.oViewKeys && this.oScreenSchema){
            for (const [dataKey, data] of Object.entries(this.oScreenSchema)) {
                if(data.viewKey){
                    if(this.oViewKeys[data.viewKey]){
                        for (const [viewKeyProp, viewPropValue] of Object.entries(this.oViewKeys[data.viewKey])) {
                            data[viewKeyProp] = viewPropValue;
                        }
                    }
                }
            } 
        }
    }

    _updateTemplateWithSchema(){    
        if(this.oScreenSchema){
            for (const [dataKey, data] of Object.entries(this.oScreenSchema)) {
                let dDataKeyElem = this._bindDataKeyData(dataKey, data, this.dScreenTemplateFragment);
                
                if(dDataKeyElem){
                    this.oScreenSchema[dataKey].dRef = dDataKeyElem;
                }                    
            }              
        }
    }

    _bindDataKeyData(dataKey, data, fragment){
        let dDataKeyElem = fragment.querySelector('[dataKey="'+dataKey+'"]');

        if(dDataKeyElem && dDataKeyElem.setAttributes){
            dDataKeyElem.setAttributes(data);                 
        }

        return dDataKeyElem;
    }   
     
    async _populateSchemaWithData(){
        // Since promises are used to retreive data from various sources, this should be chained to the rest of the template and binding methods used by the screen. 
        
        let oRegisteredSources = this.cDataInterface.getSources();
            if(oRegisteredSources){
                await this.cDataInterface.populateSchema(this.oScreenSchema).then(oPopulatedSchema => {
                
                this.oScreenSchema = oPopulatedSchema;                
            });       
        }
    }

    async _populateSchemaWithPageSchema(){
        if(this.oPageSchemaEndpoint){
            if(this.oPageSchemaEndpoint.url){
                let oPageSchemaResponse = await cAPP.fetch(this.oPageSchemaEndpoint)
                if(oPageSchemaResponse){
                    //TODO: Add response validation here.
                    this.oPageSchema = oPageSchemaResponse.result[0].body;
                    this._processPageSchema();

                }
            }
        }
    }

    _processPageSchema(){
        if(this.oPageSchema.fields){
            if(this.oScreenSchema){
                for (const [dataKey, data] of Object.entries(this.oPageSchema.fields)) {                  
                    if(this.oScreenSchema[dataKey]){
                        for (const [viewKeyProp, viewPropValue] of Object.entries(this.oPageSchema.fields[dataKey])) {
                           this.oScreenSchema[dataKey][viewKeyProp] = viewPropValue;
                        }
                    }
                    else{
                        this.oScreenSchema[dataKey] = data
                    }                   
                } 
            }
            else{
                this.oScreenSchema = this.oPageSchema.fields;
            }    
        }        
    }

    _createScreenTemplateFragment(){
        if(this.sScreenTemplate){
            // Create dom fragment of template. 
            this.dScreenTemplateFragment = document.createElement('div');    //Using a div to perserve reference and not need to recreate every time.     
            this.dScreenTemplateFragment.innerHTML = this.sScreenTemplate;
            // Bind/replace all data/labels
        }   
        else if(this.sRenderTemplate){
            this.dScreenTemplateFragment = this.cRenderer.generate(this.sRenderTemplate, false, this.oRenderData);
        }     
    }

}